%% ***********************************************************************%
% All rights reserved � to the authors: Mahmoud SHAQFA and Zoltan ORBAN   %
%                                                                         %
% M. Shaqfa Contact:                                                      %
% Earthquake Engineering and Structural Dynamics Laboratory (EESD),       %
% School of Architecture, Civil and Environmental Engineering (ENAC),     %
% Ecole polytechnique federale de Lausanne (EPFL),                        %
% CH-1015 Lausanne, Switzerland.                                          %
%               Tel.: +41 21 69 33297                                     %
%               Email: mahmoud.shaqfa@epfl.ch                             %
%                                                                         %
% Z. ORBAN Contact:                                                       %
% Department of Structural Engineering,                                   %
% Faculty of Engineering and Information Technology,                      %
% University of Pecs, Boszorkany u. 2,                                    %
% H-7624 Pecs, Hungary.                                                   %
%               Email: orbanz@mik.pte.hu                                  %
%*************************************************************************%
% - This code is just a simple demonstration for the modified algorithm   %
%   without taking into account filtering the population(HM).             %
% - This code is not optimized. A vectorized code should make it          %
%   work faster or even Object Oriented Programming (OOP).                %
%*************************************************************************%
% Main references for this paper and code:                                %
%                                                                         %
% [1] - Geem, Z.W.: Optimal cost design of water distribution             %
%       networks usingharmony search 38, 259-277(2006)                    %
%                                                                         %
% [2] - Geem, Z.W.: Novel derivative of harmony search algorithm for      %
%       discrete design variables. Applied Mathematics and Computation    %
%       199(1), 223-230(2008). DOI https://doi.org/10.1016/j.amc.2007.    %
%       09.049.                                                           %
%                                                                         %
% [3] - Geem, Z.W., Sim, K.B.: Parameter-setting-free harmony search      %
%       algorithm 217, 3881-3889 (2010)                                   %
%*************************************************************************%
% This file is free: you can redistribute it and/or modify                %
% it under the terms of the GNU General Public License as published by    %
% the Free Software Foundation, either version 3 of the License, or       %
% (at your option) any later version.                                     %
%                                                                         %
% This file is distributed in the hope that it will be useful,            %
% but WITHOUT ANY WARRANTY; without even the implied warranty of          %
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           %
% GNU General Public License for more details.                            %
% TO recieve a copy of GNU License refer to: https://www.gnu.org/licenses/%
%*************************************************************************%
function FinalResults = PSFHS_Algorithm
%% Callibration and PSFHS Parameters
clear; close all; clc; format short;
rng('shuffle') % use the current-time seed


% PSFHS Parameters Setting
MaxIt = 300000;      % Max. Number of Iterations (Stopping Criteria)
m=2.5;               % Harmony Memory Rehearsal Multiplier (HMS x m)
HMCRi= 0.75;         % Initial Harmony Memory Consideration Rate
PARi = 0.15;         % Initial Pitch Adjustment Rate
HMCR_max = 0.99;     % Max HMCR
PAR_min = 0.05;      % Min PAR
HMS= 25;             % Harmony Memory Size (The Population Size)
%% Related to the objective function (Schwefel Function)
% For more info refer to: https://www.sfu.ca/~ssurjano/schwef.html
N = 100;                               % Number of the problem's dimensions
% The exact answer is (See the objective function below):
% Exact answer = - N * 418.9829
% The Answer roots (dimensions) = X(1 -> N) = -/+ 420.9687 (for all values)
LB =  ones(1,N) * -500;                % Lower Bound for the objective
UB =  ones(1,N) * 500;                 % Upper Bound for the objective
StepSize = ones(1,N) * 0.001;          % The Accuracy of solution roots
% [1] --> Plot Results; [else] --> Ignore Plots;
AlgorithmVisualization = 1;
%% The PSFHS Process Starts Here.
% PSFHS Initialization and Pre-Allocation
HM = zeros(HMS,N);                     % Pre-Allocation Harmony Memory. (Not Recommended Pre-Allocation)
FunVal=zeros(HMS,1);                   % Pre-Allocation of Fitnesses
Rehearsal = ceil(m*HMS);               % Amount of Rehearsals m x HMS (in the Related Study) - Real Positive Integer
OM = zeros(HMS,N);                     % Rehearsal Operations Memory Associated with filling up each Solution Vector
OI = zeros(1,N);                       % Pre-Allocation of Operation Index Vector Values of [1,2 or 3]
Xnew = OI;                             % Pre-Allocation for New Vector of Variables
HMCRRunsHistory=zeros(MaxIt,N);        % HMCR for All Runs and Variables
PARRunsHistory=zeros(MaxIt,N);         % PAR  for All Runs and Variables
BestCompHistory = zeros(MaxIt, 1);     % Best History Per Iteration

%% The Core of PSFHS Algorithm
tic
% Initialize the Harmony Memory Population (HM) - Step 1 in PSFHS
for i=1:HMS
    for j=1:N
        HM(i,j)=LB(j)+(UB(j)-LB(j))*rand;
    end
    FunVal(i)= Objective(HM(i,:));
end

fprintf('\n The Modified Parameters-Setting-Free Harmony Search PSFHS Results \n');  
% Rehearsal Step in PSFHS - Step 2 in PSFHS
fprintf('\n\n The Rehearsal Step Started, The First %1.0f Iteration/s \n',Rehearsal);
HMCRRunsHistory(1:Rehearsal,:)=HMCRi;  % Assigning Rehearsal HMCR by HMCRi
PARRunsHistory(1:Rehearsal,:)=PARi;    % Assigning Rehearsal PAR by PARi
for i = 1:Rehearsal
    for j = 1:N
        if rand < HMCRi
            Xnew(j) = HM(randi(HMS),j); OI(j) = 1;
            if rand < PARi
                Xnew(j) = Xnew(j) + (-1)^randi(2) * round(rand * (UB(j) - LB(j)) / StepSize(j)) * StepSize(j);
                OI(j) = 2;
            end
        else
            Xnew(j) = LB(j) + round(rand * (UB(j) - LB(j)) / StepSize(j)) * StepSize(j);
            OI(j) = 3;
        end
        Xnew(j)=min([ max([ Xnew(j) LB(j) ]) UB(j)]);
    end
    FNew = Objective(Xnew);
    [Fworst, WorstIndex] = max(FunVal);
    if FNew < Fworst
        HM(WorstIndex,:) = Xnew;
        FunVal(WorstIndex) = FNew;
        OM(WorstIndex,:)= OI;
    end
    [CurrentBest, ~] = min(FunVal);
    BestCompHistory(i) = CurrentBest;
    fprintf('\n Itr# %1.0f, Best Answer = %5.5f',i...
            ,CurrentBest);
    OI = zeros(1,N);
end

% The Performance Step of the PSFHS - Step 3 in PSFHS
fprintf('\n\n The Performance Step Started, For the Rest %1.0f Iteration/s \n',MaxIt-Rehearsal)
for i = (Rehearsal+1):MaxIt
        for j = 1:N
            HMCR = sum(OM(:,j) == 1) / HMS;
            PAR  = sum(OM(:,j) == 2) / HMS;
            HMCR = max([HMCRi min([HMCR_max HMCR])]);
            PAR  = max([PAR_min min([PARi   PAR])]);
            HMCRRunsHistory(i,j) = HMCR;
            PARRunsHistory(i,j) = PAR;
            if rand < HMCR
                Xnew(j) = HM(randi(HMS),j); OI(j) = 1;
                if rand < PAR
                    Xnew(j) = Xnew(j) + (-1) ^ randi(2) * round(rand * (UB(j) - LB(j)) / StepSize(j)) * StepSize(j);
                    OI(j) = 2;
                end
            else
                Xnew(j) = LB(j) + round(rand*(UB(j)-LB(j))/StepSize(j))*StepSize(j);
                OI(j) = 3;
            end
            Xnew(j)=min([ max([ Xnew(j) LB(j) ]) UB(j)]);
        end
        FNew = Objective(Xnew);
        [Fworst, WorstIndex] = max(FunVal);
        if FNew < Fworst
            HM(WorstIndex,:) = Xnew;
            FunVal(WorstIndex) = FNew;
            OM(WorstIndex,:)= OI;
        end
        [CurrentBest, ~] = min(FunVal);
        BestCompHistory(i) = CurrentBest;
        fprintf('\n Itr# %1.0f, Best Answer = %5.5f',i...
                ,CurrentBest);
        OI = zeros(1,N);
end

% Ranking and Choosing The Best Answer
[Best, index] = min(FunVal);
fprintf('\n\nThe Best Answer = %1.10f ,Where:',Best);
for i=1:N
	eval(['X' num2str(i) '=HM(index,i)']);
end
fprintf('\nLast Convergance Update, At Iteration No. %1.0f\n'...
	,lastupdate(BestCompHistory, MaxIt));
BestHistory = Best;
BestVectorHistory = HM(index,:);
%% Data Visualization and Results Reporting
if AlgorithmVisualization == 1
    figure
    semilogx(1:MaxIt, BestCompHistory,'LineWidth',1.7);
    title('Modified PSFHS Benchmarking Example')
    xlabel('Iterations');
    ylabel('Function Evaluation');
    legend('PSFHS - Modified Algorithm');
    xlim([1 MaxIt]);
    grid on
    figure
    for j = 1:N
        subplot(2,1,1)
        ZZ = rand(1,3); % Color Vector
        semilogx(1:MaxIt,HMCRRunsHistory(:,j),'color',ZZ,'LineWidth',1.7);
        title('History of HMCR');
        xlabel('Iterations')
        ylabel('Harmony Memory Consideration Rate (HMCR)')
        xlim([1 MaxIt]);grid on;
        hold on
        subplot(2,1,2)
        semilogx(1:MaxIt,PARRunsHistory(:,j),'color',ZZ,'LineWidth',1.7);
        title('History of PAR');
        xlabel('Iterations')
        ylabel('Pitch Adjustment Rate (PAR)')
        xlim([1 MaxIt]);grid on;
        hold on
        legendInfo{j} = ['X ' num2str(j)];               
    end
    legend(legendInfo);
end
 
%% Hold The Final Memory (Results)
FinalResults.RunTime = toc; % Time of Implementations in Seconds
FinalResults.HM = HM;
FinalResults.MaxIt = MaxIt;
FinalResults.BestCompHistory = BestCompHistory;
FinalResults.BestVectorHistory = BestVectorHistory;
FinalResults.BestAnswer = BestHistory;
FinalResults.OM = OM;
FinalResults.HM = HM;
end
%% Last Updates
function Counter=lastupdate(BestComp, MaxIt)
[x,~] = min(BestComp);
num = sum(BestComp==x);
Counter = MaxIt - num + 1;
end
%% Objective Functions

function obj = Objective(SolutionVector)
% See: http://benchmarkfcns.xyz/benchmarkfcns/schwefelfcn.html
obj = sum(SolutionVector .* sin(sqrt(abs(SolutionVector)))); % vectorized
% The exact answer of this form will be 418.9829*(Dims.)
% This can be modified (The objective) to be in the form:
% Obj = Obj + 418.9829 - SolutionVector(i)*sin(sqrt(abs(SolutionVector(i))))
% in this case the exact answer should be --> Zero, rather than
% -418.9829*Dims
% To validate the other case uncomment the next line:
% obj = obj + length(SolutionVector)*418.9829;
end
